# coding=utf-8
import struct
import unittest

from pamqp import constants, exceptions, frame


class TestCase(unittest.TestCase):

    def test_invalid_protocol_header(self):
        with self.assertRaises(exceptions.UnmarshalingException) as err:
            frame.unmarshal(b'AMQP\x00\x00\t')
            self.assertTrue(str(err).startswith(
                "Could not unmarshal <class 'pamqp.header.ProtocolHeader'> "
                'frame: Data did not match the ProtocolHeader format'))

    def test_invalid_frame_header(self):
        frame_data = struct.pack('>BI', 255, 0)
        with self.assertRaises(exceptions.UnmarshalingException) as err:
            frame.unmarshal(frame_data)
            self.assertEqual(
                str(err), 'Could not unmarshal Unknown frame: No frame size')

    def test_frame_with_no_length(self):
        frame_data = (b'\x01\x00\x01\x00\x00\x00\x00\x00<\x00P\x00\x00\x00\x00'
                      b'\x00\x00\x00\x01\x00\xce')
        with self.assertRaises(exceptions.UnmarshalingException) as err:
            frame.unmarshal(frame_data)
            self.assertEqual(
                str(err), 'Could not unmarshal Unknown frame: No frame size')

    def test_frame_malformed_length(self):
        frame_data = (b'\x01\x00\x01\x00\x00\x00\x0c\x00<\x00P\x00\x00\x00\x00'
                      b'\x00\x00\x00\xce')
        with self.assertRaises(exceptions.UnmarshalingException) as err:
            frame.unmarshal(frame_data)
            self.assertEqual(
                str(err),
                'Could not unmarshal Unknown frame: Not all data received')

    def test_frame_malformed_end_byte(self):
        frame_data = (b'\x01\x00\x01\x00\x00\x00\r\x00<\x00P\x00\x00\x00\x00'
                      b'\x00\x00\x00\x01\x00\x00')
        with self.assertRaises(exceptions.UnmarshalingException) as err:
            frame.unmarshal(frame_data)
            self.assertEqual(
                str(err),
                'Could not unmarshal Unknown frame: Last byte error')

    def test_malformed_frame_content(self):
        payload = struct.pack('>HxxQ', 8192, 32768)
        frame_value = b''.join([struct.pack('>BHI', 5, 0, len(payload)),
                                payload, constants.FRAME_END_CHAR])
        with self.assertRaises(exceptions.UnmarshalingException) as err:
            frame.unmarshal(frame_value)
            self.assertEqual(
                str(err),
                'Could not unmarshal Unknown frame: Unknown frame type: 5')

    def test_invalid_method_frame_index(self):
        payload = struct.pack('>L', 42949)
        frame_value = b''.join([struct.pack('>BHI', 1, 0, len(payload)),
                                payload, constants.FRAME_END_CHAR])
        with self.assertRaises(exceptions.UnmarshalingException) as err:
            frame.unmarshal(frame_value)
            self.assertEqual(
                str(err),
                ('Could not unmarshal Unknown frame: '
                 'Unknown method index: 42949'))

    def test_invalid_method_frame_content(self):
        payload = struct.pack('>L', 0x000A0029)
        frame_value = b''.join([struct.pack('>BHI', 1, 0, len(payload)),
                                payload, constants.FRAME_END_CHAR])
        with self.assertRaises(exceptions.UnmarshalingException) as err:
            frame.unmarshal(frame_value)
            self.assertTrue(str(err).startswith(
                'Could not unmarshal <pamqp.specification.Connection.OpenOk'))

    def test_invalid_content_header_frame(self):
        payload = struct.pack('>L', 0x000A0029)
        frame_value = b''.join([struct.pack('>BHI', 2, 0, len(payload)),
                                payload, constants.FRAME_END_CHAR])
        with self.assertRaises(exceptions.UnmarshalingException) as err:
            frame.unmarshal(frame_value)
            self.assertTrue(str(err).startswith(
                'Could not unmarshal ContentHeader frame:'))



#!/usr/bin/env python
"""Generates the pamqp/specification.py file used as a foundation for AMQP
communication.

"""
import copy
import dataclasses
import functools
import json
import keyword
import logging
import pathlib
import sys
import textwrap
import typing

import lxml.etree, lxml.objectify
import requests
from yapf import yapf_api

__author__ = 'Gavin M. Roy'
__email__ = 'gavinmroy@gmail.com'
__since__ = '2011-03-31'

LOGGER = logging.getLogger(__name__)

CODEGEN_DIR = pathlib.Path('./codegen/')
CODEGEN_IGNORE_CLASSES = ['access']
CODEGEN_JSON = CODEGEN_DIR / 'amqp-rabbitmq-0.9.1.json'
CODEGEN_XML = CODEGEN_DIR / 'amqp0-9-1.xml'
EXTENSIONS_XML = CODEGEN_DIR / 'extensions.xml'

COMMANDS = pathlib.Path('./pamqp/commands.py')
COMMANDS_HEADER = '''"""
The classes inside :mod:`pamqp.commands` allow for the automatic marshaling
and unmarshaling of AMQP method frames and
:class:`Basic.Properties <pamqp.commands.Basic.Properties>`. In addition the
command classes contain information that designates if they are synchronous
commands and if so, what the expected responses are. Each commands arguments
are detailed in the class and are listed in the attributes property.

.. note:: All AMQ classes and methods extend :class:`pamqp.base.Frame`.

"""
# Auto-generated, do not edit this file.
import datetime
import typing
import warnings

from pamqp import base, common, constants

'''

CONSTANTS = pathlib.Path('./pamqp/constants.py')
CONSTANTS_HEADER = '''
# Auto-generated, do not edit this file.
import re
'''

EXCEPTIONS = pathlib.Path('./pamqp/exceptions.py')
EXCEPTIONS_HEADER = '''
# Auto-generated, do not edit this file.


class PAMQPException(Exception):
    """Base exception for all pamqp specific exceptions."""


class UnmarshalingException(PAMQPException):
    """Raised when a frame is not able to be unmarshaled."""
    def __str__(self) -> str:  # pragma: nocover
        return 'Could not unmarshal {} frame: {}'.format(
            self.args[0], self.args[1])


class AMQPError(PAMQPException):
    """Base exception for all AMQP errors."""


class AMQPSoftError(AMQPError):
    """Base exception for all AMQP soft errors."""


class AMQPHardError(AMQPError):
    """Base exception for all AMQP hard errors."""

'''

CODEGEN_JSON_URL = ('https://raw.githubusercontent.com/rabbitmq/'
                    'rabbitmq-codegen/master/amqp-rabbitmq-0.9.1.json')
CODEGEN_XML_URL = ('https://raw.githubusercontent.com/rabbitmq/'
                   'rabbitmq-website/master/site/resources/specs/'
                   'amqp0-9-1.extended.xml')

XPATH_ORDER = ['class', 'constant', 'method', 'field']

AMQ_TYPE_TO_ANNOTATION = {
    'bit': 'bool',
    'long': 'int',
    'longlong': 'int',
    'longstr': 'str',
    'octet': 'int',
    'short': 'int',
    'shortstr': 'str',
    'table': 'common.FieldTable',
    'timestamp': 'common.Timestamp',
}


@dataclasses.dataclass
class Domain:
    name: str
    type: str
    documentation: typing.Optional[str]
    label: typing.Optional[str]
    nullable: bool
    regex: typing.Optional[str]
    max_length: typing.Optional[int]
    default_value: typing.Optional[str]


class Codegen:

    AMQP_VERSION = '0-9-1'
    DEPRECATION_WARNING = 'Deprecated'
    YAPF = True

    def __init__(self):
        self._amqp_json = self._load_codegen_json()
        self._codegen_xml = self._load_codegen_xml()
        self._extensions_xml = self._load_extensions_xml()
        self._output_buffer: typing.List[str] = []
        self.AMQP_VERSION = ('-'.join([
            str(self._amqp_json['major-version']),
            str(self._amqp_json['minor-version']),
            str(self._amqp_json['revision'])]))
        self.DEPRECATION_WARNING = \
            'This command is deprecated in AMQP {}'.format(self.AMQP_VERSION)

    def build(self):
        self._build_commands()
        self._build_constants()
        self._build_exceptions()

    def _add_comment(self, value: str, indent: int = 0,
                     prefix: str = '# ') -> None:
        """Append a comment to the output buffer"""
        if not value:
            return self._add_line()
        indent_text = prefix.rjust(indent + len(prefix))
        wrapped = textwrap.wrap(value, 79 - len(indent_text))
        if value.startswith(':') or value.startswith('..'):
            self._add_line(indent_text + wrapped[0], width=79)
            indent += 4
            indent_text = prefix.rjust(indent + len(prefix))
            value = value[len(wrapped[0]) + 1:]
            for line in  textwrap.wrap(value, 79 - len(indent_text)):
                self._add_line(indent_text + line, width=79)
        else:
            for line in wrapped:
                self._add_line(indent_text + line, width=79)

    def _add_documentation(self, label: str, documentation: str,
                           indent: int) -> None:
        if not documentation:
            self._add_line('"""{}"""'.format(label), indent)
        else:
            for key, value in sorted(self._commands().items(), reverse=True):
                documentation = documentation.replace(
                    ' {} '.format(key), ' :class:`{}` '.format(value))
            self._add_line('"""{}'.format(label), indent)
            self._add_line()
            for line in documentation.split('\n'):
                self._add_comment(line, indent, '')
            self._add_line()
            self._add_line('"""', indent)

    def _add_function(self, name: str, args: list, indent: int) -> None:
        """Create a new function"""
        self._add_line()
        if not len(args):
            self._add_line('def {}(self) -> None:'.format(name), indent)
        else:
            self._add_line('def {}(self,'.format(name), indent)
            indent += len('def {}('.format(name))
            for index, arg in enumerate(args):
                annotation = self._arg_annotation(arg)
                default = self._arg_default(arg)
                if arg['name'] == 'arguments' and default == '{}':
                    default = 'None'
                if default == 'None' or default is None:
                    annotation = 'typing.Optional[{}]'.format(annotation)
                if index == len(args) - 1:
                    self._add_line('{}: {} = {}) -> None:'.format(
                        self._arg_name(arg['pyname']),
                        annotation, default), indent)
                else:
                    self._add_line('{}: {} = {},'.format(
                        self._arg_name(arg['pyname']),
                        annotation, default), indent)

    def _add_line(self, value: str = '', indent: int = 0,
                  secondary_indent: int = 0,
                  width: typing.Optional[int] = None) -> None:
        """Append a new line to the output buffer"""
        if not value:
            self._output_buffer.append(value)
            return
        initial = ''.rjust(indent)
        secondary = ''.rjust(secondary_indent or indent)
        wrapper = textwrap.TextWrapper(
            width=width or 120 if self.YAPF else 79,
            drop_whitespace=True,
            initial_indent=initial,
            break_on_hyphens=False,
            subsequent_indent=secondary)

        for value in wrapper.wrap(value.rstrip()):
            self._output_buffer.append(value)

    def _all_defaults(self, arguments: list) -> bool:
        args = []
        for arg in arguments:
            arg_type = AMQ_TYPE_TO_ANNOTATION[self._arg_type(arg)]
            if arg_type in {'common.FieldTable', 'common.Arguments'}:
                args.append(True)
            else:
                args.append(self._arg_default(arg) is not None)
        return all(args)

    def _arg_annotation(self, arg: dict) -> str:
        arg_type = AMQ_TYPE_TO_ANNOTATION[self._arg_type(arg)]
        if arg_type == 'common.FieldTable':
            if arg['name'] == 'arguments':
                return 'common.Arguments'
        elif arg_type == 'common.Timestamp':
            return 'datetime.datetime'
        return arg_type

    def _arg_default(self, arg: dict, repr_=True) -> typing.Optional[str]:
        ext = self._extensions_xml.xpath(
            '//rabbitmq/field[@name="{}"]'.format(arg['name']))
        if ext and ext[0].attrib.get('default-value') is not None:
            if repr_ and arg['type'] != 'short' and arg['type'] != 'bit':
                return '{!r}'.format(ext[0].attrib.get('default-value'))
            return ext[0].attrib.get('default-value')
        elif arg.get('default-value') is not None:
            if repr_ and arg['type'] != 'short' and arg['type'] != 'bit':
                return '{!r}'.format(arg['default-value'])
            return arg['default-value']
        elif 'domain' in arg:
            domain = self._domain(arg['domain'])
            if domain and domain.name == arg['domain'] \
                    and domain.default_value is not None:
                if repr_ and domain.type != 'short' and arg['type'] != 'bit':
                    return '{!r}'.format(domain.default_value)
                return domain.default_value
        return None

    @staticmethod
    def _arg_name(name: str) -> str:
        """Returns a valid python argument name for the AMQP argument"""
        value = name.replace('-', '_')
        if value in keyword.kwlist:
            LOGGER.debug('%s is in the keyword list', value)
            value += '_'
        return value

    def _arg_type(self, arg: dict) -> str:
        """Get the argument type"""
        if 'domain' in arg:
            for domain, data_type in self._amqp_json['domains']:
                if arg['domain'] == domain:
                    arg['type'] = data_type
                    break
        if 'type' in arg:
            return arg['type']
        raise ValueError('Unknown argument type')

    def _build_command_basic_properties(self, class_name: str,
                                        class_id: int,
                                        properties: list) -> None:
        for offset, arg in enumerate(properties):
            properties[offset]['pyname'] = self._arg_name(arg['name'])
            if properties[offset]['pyname'] == 'type':
                properties[offset]['pyname'] = 'message_type'

        indent = 4
        self._add_line('class Properties(base.BasicProperties):', indent)
        indent += 4
        self._add_comment('"""Content Properties', indent, '')
        self._add_line()
        self._add_line(
            '.. Note:: The AMQP property type is named ``message_type`` as to '
            'not conflict with the Python ``type`` keyword',
            indent, indent + 10, width=79)
        self._add_line()
        for arg in properties:
            label = self._label({'class': class_name, 'field': arg['name']})
            if label:
                line = ':param {}: {}'.format(arg['pyname'], label or '')
                self._add_line(line.strip(), indent, indent + 4, width=79)
            else:
                self._add_line(':param {}: Deprecated, must be empty'.format(
                    arg['pyname']), indent, indent + 4, width=79)
            if arg['name'] == 'headers':
                self._add_line(
                    ':type headers: typing.Optional['
                    ':const:`~pamqp.common.FieldTable`]',
                    indent, indent + 4, width=79)
        self._add_line(':raises: ValueError', indent)
        self._add_line()
        self._add_line('"""', indent)

        self._add_line('__annotations__: typing.Dict[str, object] = {',
                       indent)
        for offset, arg in enumerate(properties):
            if offset == len(properties) - 1:
                self._add_line(
                    '{!r}: {}'.format(arg['pyname'], self._arg_annotation(arg)),
                    indent + 4)
            else:
                self._add_line(
                    '{!r}: {},'.format(arg['pyname'], self._arg_annotation(arg)),
                    indent + 4)
        self._add_line('}', indent)

        self._add_line(
            '__slots__: typing.List[str] = [  # AMQ Properties Attributes',
            indent)
        for offset, arg in enumerate(properties):
            if offset == len(properties) - 1:
                self._add_line('{!r}'.format(arg['pyname']), indent + 4)
            else:
                self._add_line('{!r},'.format(arg['pyname']), indent + 4)
        self._add_line(']', indent)
        self._add_line()

        flag_value = 15
        self._add_comment('Flag values for marshaling / unmarshaling', indent)
        self._add_line(
            "flags = {{'{}': {},".format(
                properties[0]['pyname'], 1 << flag_value), indent)
        for arg in properties[1:-1]:
            flag_value -= 1
            self._add_line("'{}': {},".format(arg['pyname'], 1 << flag_value),
                           indent + 9),
        flag_value -= 1
        self._add_line("'{}': {}}}".format(
            properties[-1]['pyname'], 1 << flag_value),
            indent + 9)
        self._add_line()
        self._add_line('frame_id = {}  # AMQP Frame ID'.format(class_id),
                       indent)
        self._add_line(
            'index = 0x%04X  # pamqp Mapping Index' % class_id, indent)
        self._add_line("name = '{}.Properties'".format(
            self._pep8_class_name(class_name)), indent)
        self._add_line()

        self._add_comment('Class Attribute Types for unmarshaling', indent)
        for arg in properties:
            self._add_line("_{} = '{}'".format(
                arg['pyname'], self._arg_type(arg)), indent)

        self._add_function('__init__', properties, indent)
        indent += 4
        self._add_line('"""Initialize the {}.Properties class"""'.format(
            self._pep8_class_name(class_name)), indent)
        for arg in properties:
            self._add_line('self.{} = {}'.format(
                arg['pyname'], arg['pyname']), indent)
        self._add_line('self.validate()', indent)
        self._add_line()

    def _build_command_map(self) -> None:
        self._add_line()
        self._add_comment('AMQP Class.Method Index Mapping')
        self._add_line('INDEX_MAPPING = {')
        lines = []
        for amqp_class in self._amqp_json['classes']:
            if amqp_class['name'] not in CODEGEN_IGNORE_CLASSES:
                for method in amqp_class['methods']:
                    key = amqp_class['id'] << 16 | method['id']
                    lines.append(
                        '0x%08X: %s.%s,' % (
                            key,
                            self._pep8_class_name(amqp_class['name']),
                            self._pep8_class_name(method['name'])))
        lines[-1] = lines[-1][:-1]
        [self._add_line(line, 4) for line in lines]
        self._add_line('}')

    def _build_command_method(self, class_name: str, class_id: int,
                              method: dict) -> None:
        deprecated, needs_linefeed = False, True
        arguments = copy.deepcopy(method['arguments'])
        method_xml = self._search_xml(
            {'class': class_name, 'method': method['name']})
        indent = 4
        self._add_line('class {}(base.Frame):'.format(
            self._pep8_class_name(method['name'])), indent)
        documentation = []
        docs = self._documentation(
            {'class': class_name, 'method': method['name']})
        if docs is not None and docs.strip():
            documentation.append(docs.strip())
            needs_linefeed = True

        for arg in arguments:
            name = self._arg_name(arg['name'])
            if name == 'type' and class_name == 'exchange':
                documentation.append(
                    '\n.. note:: The AMQP type argument is referred to as '
                    '"{}_type" to not conflict with the Python type '
                    'keyword.'.format(class_name))
                needs_linefeed = True

        # Note the deprecation warning in the docblock
        if method_xml and method_xml[0].attrib.get('deprecated'):
            deprecated = True
            documentation.append(
                '\n.. deprecated:: {}'.format(self.DEPRECATION_WARNING))
            needs_linefeed = True

        if arguments and needs_linefeed:
            documentation.append('')

        for arg in arguments:
            arg_type = AMQ_TYPE_TO_ANNOTATION[self._arg_type(arg)]
            name = self._arg_name(arg['name'])
            default = self._arg_default(arg, False)
            if name == 'arguments' or arg_type in {'common.FieldTable',
                                                   'common.Arguments'}:
                default = '{}'
            domain = self._domain(arg.get('domain', arg['name']))
            if name == 'type' and class_name == 'exchange':
                name = 'exchange_type'
            search_path = {
                'class': class_name,
                'method': method['name'],
                'field': arg['name']}
            arg_doc = self._documentation(search_path)
            label = self._label(search_path) or arg_doc or ''
            if not label \
                    and domain and (domain.label or domain.documentation):
                label = domain.label or domain.documentation
            if default is not None:
                if default == '':
                    default = "''"
                if label:
                    label = '{}\n    - Default: ``{}``'.format(label, default)
                else:
                    label = 'Default: ``{}``'.format(default)
            documentation.append(':param {}: {}'.format(name,  label))
            if name == 'arguments':
                documentation.append(
                    ':type {}: :const:`~pamqp.common.Arguments`'.format(name))
            elif arg_type.startswith('common'):
                documentation.append(
                    ':type {}: :const:`~pamqp.{}`'.format(name, arg_type))

        raises = set()
        for arg in arguments:
            domain = self._domain(arg.get('domain', arg['name']))
            if domain and (domain.max_length or domain.regex):
                raises.add('ValueError')
        if raises:
            for exc_name in raises:
                documentation.append(
                    ':raises {}: when an argument fails to validate'.format(
                        exc_name))

        label = self._label(
            {'class': class_name, 'method': method['name']}) \
            or 'Undocumented function'

        indent = 8
        self._add_documentation(label, '\n'.join(documentation), indent)

        if not len(arguments):
            self._add_line('__annotations__: typing.Dict[str, object] = {}',
                           indent)
        else:
            self._add_line('__annotations__: typing.Dict[str, object] = {',
                           indent)
            for offset, arg in enumerate(arguments):
                name = self._arg_name(arg['name'])
                if name == 'type' and class_name == 'exchange':
                    name = 'exchange_type'
                arguments[offset]['pyname'] = name
                if offset == len(method['arguments']) - 1:
                    self._add_line(
                        '{!r}: {}'.format(name, self._arg_annotation(arg)),
                        indent + 4)
                else:
                    self._add_line(
                        '{!r}: {},'.format(name, self._arg_annotation(arg)),
                        indent + 4)
            self._add_line('}', indent)
        if not len(arguments):
            self._add_line(
                '__slots__: typing.List[str] = []  # AMQ Method Attributes',
                indent)
        else:
            self._add_line(
                '__slots__: typing.List[str] = [  # AMQ Method Attributes',
                indent)
            for offset, arg in enumerate(arguments):
                if offset == len(method['arguments']) - 1:
                    self._add_line('{!r}'.format(arg['pyname']), indent + 4)
                else:
                    self._add_line('{!r},'.format(arg['pyname']), indent + 4)
            self._add_line(']', indent)
        self._add_line()
        self._add_line('frame_id = %i  # AMQP Frame ID' % method['id'], indent)
        index_value = class_id << 16 | method['id']
        self._add_line(
            'index = 0x%08X  # pamqp Mapping Index' % index_value, indent)
        self._add_line("name = '{}.{}'".format(
            self._pep8_class_name(class_name),
            self._pep8_class_name(method['name'])), indent)
        self._add_line(
            'synchronous = {}  # Indicates if this is a synchronous AMQP '
            'method'.format(method.get('synchronous', False)), indent)

        if method.get('synchronous'):
            responses = []
            if method_xml:
                responses = [
                    "'{}.{}'".format(
                        self._pep8_class_name(class_name),
                        self._pep8_class_name(response.attrib['name']))
                    for response in method_xml[0].iter('response')]
            if not responses:
                responses = ["'{}.{}Ok'".format(
                    self._pep8_class_name(class_name),
                    self._pep8_class_name(method['name']))]
            line = 'valid_responses = [{}]'.format(', '.join(responses))
            if len(line) <= 36:
                self._add_line('{}  # Valid responses to this method'.format(
                    line), indent)
            else:
                self._add_comment('Valid responses to this method', indent)
                self._add_line(line, indent)

        if arguments:
            self._add_line()
            self._add_comment('Class Attribute Types for unmarshaling', indent)
        for arg in arguments:
            self._add_line("_{} = '{}'".format(arg['pyname'], arg['type']),
                           indent)
        if len(arguments):
            self._add_function('__init__', arguments, indent)
            indent += 4
            self._add_line(
                '"""Initialize the :class:`{}.{}` class"""'.format(
                    self._pep8_class_name(class_name),
                    self._pep8_class_name(method['name'])), indent)

            # Function
            all_defaults = self._all_defaults(arguments)
            for arg in arguments:
                domain = self._domain(arg.get('domain', arg['name']))
                default = self._arg_default(arg)
                if ((domain and not domain.nullable and domain.type == 'table')
                        or arg['name'] == 'arguments'
                        or (arg['type'] == 'table' and not default)):
                    self._add_line('self.{} = {} or {{}}'.format(
                        arg['pyname'], arg['pyname']), indent)
                elif default is not None \
                        and not all_defaults \
                        and default != '':
                    self._add_line('self.{} = {} or {}'.format(
                        arg['pyname'], arg['pyname'], default), indent)
                else:
                    self._add_line('self.{} = {}'.format(
                        arg['pyname'], arg['pyname']), indent)

            if deprecated:
                self._add_line('warnings.warn(', indent)
                self._add_line(
                    'constants.DEPRECATION_WARNING, '
                    'category=DeprecationWarning)', indent + 4)

            add_validate = False
            for arg in arguments:
                domain = self._domain(arg.get('domain', arg['name']))
                if ((arg['pyname'] in {'capabilities', 'channel_id',
                                       'cluster_id', 'known_hosts',
                                       'out_of_band', 'internal',
                                       'insist', 'ticket'})
                    or domain and domain.max_length is not None
                    or domain and domain.regex is not None):
                    add_validate = True
                    break

            if add_validate:
                self._add_line('self.validate()', indent)

            indent -=4

            def format_deprecated_value(
                    field_name: str,
                    operator: str,
                    expected_value: typing.Union[str, bool, int],
                    error_value: str) -> None:
                self._add_line(
                    "if self.{} is not None and self.{} {} {}:".format(
                        field_name, field_name, operator, expected_value),
                    indent)
                self._add_line(
                    "raise ValueError('{} must be {}')".format(
                        field_name, error_value), indent + 4)

            if add_validate:
                self._add_function('validate', [], indent)
                indent += 4
                self._add_comment(
                    '"""Validate the frame data ensuring all domains or '
                    'attributes adhere to the protocol specification.',
                    indent, '')
                self._add_line()
                self._add_line(':raises ValueError: on validation error', indent)
                self._add_line()
                self._add_line('"""', indent)
                for arg in arguments:
                    domain = self._domain(arg.get('domain', arg['name']))
                    if arg['pyname'] in {'capabilities', 'cluster_id',
                                         'known_hosts'}:
                        format_deprecated_value(
                            arg['pyname'], '!=', "''", 'empty')
                        continue
                    elif arg['name'] in {'internal', 'insist'}:
                        format_deprecated_value(arg['pyname'], 'is not',
                                                'False', 'False')
                        continue
                    elif arg['pyname'] == 'ticket':
                        format_deprecated_value(arg['pyname'], '!=', '0', '0')
                        continue
                    elif arg['pyname'] in {'channel_id', 'out_of_band'}:
                        format_deprecated_value(
                            arg['pyname'], '!=', "'0'", '0')
                        continue
                    if domain and domain.max_length is not None:
                        self._add_line(
                            'if self.{} is not None and '
                            'len(self.{}) > {}:'.format(
                                arg['pyname'], arg['pyname'],
                                domain.max_length),
                            indent)
                        line = ("raise ValueError('Max length "
                                "exceeded for {}')".format(arg['pyname']))
                        self._add_line(line, indent + 4)
                    if domain and domain.regex is not None:
                        self._add_line(
                            'if self.{} is not None and '
                            "not constants.DOMAIN_REGEX['{}']"
                            ".fullmatch(self.{}):".format(
                                arg['pyname'], domain.name, arg['pyname']),
                            indent)
                        line = 'raise ValueError(' \
                               "'Invalid value for {}')".format(arg['pyname'])
                        self._add_line(line, indent + 4)

        self._add_line()

    def _build_commands(self):
        LOGGER.info('Generating %s', COMMANDS)
        self._output_buffer = [COMMANDS_HEADER]
        for name in [c['name'] for c in self._amqp_json['classes']
                     if c['name'] not in CODEGEN_IGNORE_CLASSES]:
            indent = 4
            definition = self._class_definition(name)
            self._add_line()
            self._add_line('class %s:' % self._pep8_class_name(name))

            documentation = self._documentation({'class': name})
            label = self._label({'class': name}) or 'Undefined label'
            if documentation:
                self._add_documentation(label, documentation, indent)

            self._add_line('__slots__: typing.List[str] = []', indent)
            self._add_line()
            self._add_line('frame_id = {}  # AMQP Frame ID'.format(
                definition['id']), indent)
            self._add_line('index = 0x%08X  # pamqp Mapping Index' % (
                definition['id'] << 16), indent)
            self._add_line()

            for method_defn in definition['methods']:
                self._build_command_method(name, definition['id'], method_defn)

            if 'properties' in definition and definition['properties']:
                self._build_command_basic_properties(
                    name, definition['id'], definition['properties'])

        self._build_command_map()
        self._write_file(COMMANDS)

    def _build_constants(self):
        LOGGER.info('Generating %s', CONSTANTS)

        def _build_domain_output() -> typing.Tuple[typing.List[str],
                                                   typing.List[str],
                                                   typing.List[str]]:
            data_types, domains, dom_regex = [], [], []
            for domain in self._domains():
                if domain.name == domain.type:
                    value = "              '{}',".format(domain.name)
                    comment = '{} {}'.format('#'.rjust(28 - len(value)),
                                             domain.label)
                    data_types.append('{}{}'.format(value, comment))
                else:
                    domains.append("           '{}': '{}',".format(
                        domain.name, domain.type))
                    if domain.regex:
                        dom_regex.append(
                            "           '{}': re.compile(r'{}'),".format(
                                domain.name, domain.regex))

            data_types[0] = data_types[0].replace(
                '              ', 'DATA_TYPES = [')
            data_types[-1] = data_types[-1].replace(',', ' ')
            data_types.append(']')
            domains[0] = domains[0].replace('           ', 'DOMAINS = {')
            domains[-1] = domains[-1].replace(',', '}')

            dom_regex[0] = dom_regex[0].replace(
                '           ', 'DOMAIN_REGEX = {')
            dom_regex[-1] = dom_regex[-1].replace(',', '}')
            return data_types, domains, dom_regex

        self._output_buffer = [CONSTANTS_HEADER]
        self._add_comment('AMQP Protocol Frame Prefix')
        self._add_line("AMQP = b'AMQP'")
        self._add_line()

        self._add_comment('AMQP Protocol Version')
        self._add_line('VERSION = ({}, {}, {})'.format(
            self._amqp_json['major-version'],
            self._amqp_json['minor-version'],
            self._amqp_json['revision']))
        self._add_line()

        # Defaults
        self._add_comment('RabbitMQ Defaults')
        self._add_line("DEFAULT_HOST = 'localhost'")
        self._add_line('DEFAULT_PORT = {}'.format(self._amqp_json['port']))
        self._add_line("DEFAULT_USER = 'guest'")
        self._add_line("DEFAULT_PASS = 'guest'")
        self._add_line("DEFAULT_VHOST = '/'")
        self._add_line()

        # Constant
        self._add_comment('AMQP Constants')
        for constant in self._amqp_json['constants']:
            if 'class' not in constant:
                documentation = self._documentation(
                    {'constant': constant['name'].lower()})
                if documentation:
                    self._add_comment(documentation)
                self._add_line('{} = {}'.format(
                    self._dashify(constant['name']), constant['value']))
        self._add_line()

        self._add_comment('Not included in the spec XML or JSON files.')
        self._add_line("FRAME_END_CHAR = b'\\xce'")
        self._add_line('FRAME_HEADER_SIZE = 7')
        self._add_line('FRAME_MAX_SIZE = 131072')
        self._add_line()

        data_type_output, domain_output, domain_regex = _build_domain_output()

        self._add_comment('AMQP data types')
        self._output_buffer += data_type_output
        self._add_line()

        self._add_comment('AMQP domains')
        self._output_buffer += domain_output
        self._add_line()

        self._add_comment('AMQP domain patterns')
        self._output_buffer += domain_regex
        self._add_line()

        self._add_comment('Other constants')
        self._add_line("DEPRECATION_WARNING = '{}'".format(
            self.DEPRECATION_WARNING))
        self._add_line()

        self._write_file(CONSTANTS)

    def _build_exceptions(self):
        LOGGER.info('Generating %s', EXCEPTIONS)

        self._output_buffer = [EXCEPTIONS_HEADER]
        errors = {}
        for constant in self._amqp_json['constants']:
            if 'class' in constant:
                class_name = self._classify(constant['name'])
                if constant['class'] == 'soft-error':
                    extends = 'AMQPSoftError'
                elif constant['class'] == 'hard-error':
                    extends = 'AMQPHardError'
                else:
                    raise ValueError('Unexpected class: %s', constant['class'])
                self._add_line('class AMQP{}({}):'.format(class_name, extends))
                self._add_line('    """')
                documentation = self._documentation(
                    {'constant': constant['name'].lower()})
                if documentation:
                    self._add_comment(documentation, 4, '')
                else:
                    if extends == 'AMQPSoftError':
                        self._add_line('    Undocumented AMQP Soft Error')
                    else:
                        self._add_line('    Undocumented AMQP Hard Error')
                self._add_line()
                self._add_line('    """')
                self._add_line("    name = '%s'" % constant['name'])
                self._add_line('    value = %i' % constant['value'])
                self._add_line()
                self._add_line()
                errors[constant['value']] = class_name

        error_lines = []
        for error_code in errors.keys():
            error_lines.append(
                '          {}: AMQP{},'.format(error_code, errors[error_code]))
        self._add_comment('AMQP Error code to class mapping')
        error_lines[0] = error_lines[0].replace(
            '          ', 'CLASS_MAPPING = {')
        error_lines[-1] = error_lines[-1].replace(',', '}')
        self._output_buffer += error_lines

        self._write_file(EXCEPTIONS)

    def _class_definition(self, name: str) -> dict:
        """Iterates through classes trying to match the name against what was
        passed in.

        """
        for definition in self._amqp_json['classes']:
            if definition['name'] == name:
                for method in definition['methods']:
                    for index, ar in enumerate(method['arguments']):
                        method['arguments'][index].update(
                            self._lookup_field(
                                name, method['name'], ar['name']))
                return definition

    @staticmethod
    def _classify(name: str) -> str:
        """Replace the AMQP constant with a more pythonic classname"""
        return ''.join([part.title() for part in name.split('-')])

    @functools.lru_cache()
    def _commands(self) -> typing.List[str]:
        """Create a list of commands to replace in documentation with links
        to the appropriate classes.

        """
        commands = {'Basic.Properties': 'Basic.Properties'}
        for name in set([c['name'] for c in self._amqp_json['classes']
                         if c['name'] not in CODEGEN_IGNORE_CLASSES]):
            definition = self._class_definition(name)
            for method in definition['methods']:
                method_name = '-'.join(
                    [part.title() for part in method['name'].split('-')])
                command = '{}.{}'.format(name.title(), method_name)
                if '-' in command:
                    commands[command] = command.replace('-', '')
                else:
                    commands[command] = command
        return commands

    @staticmethod
    def _dashify(value: str) -> str:
        """Replace ``-`` with ``_`` for the passed in text"""
        return value.replace('-', '_')

    def _documentation(self, search_path: dict) -> typing.Optional[str]:
        """Find the documentation in the xpath"""
        def strip_whitespace(value: typing.List[lxml.etree.Element]) -> str:
            return ' '.join(
                [dl.strip() for dl in value[0].text.split('\n')]).strip()

        node = self._search_xml(search_path, 'doc')
        if node:
            return strip_whitespace(node)
        elif 'field' in search_path:  # Look elsewhere
            domain = self._domain(search_path['field'])
            if domain and domain.documentation:
                return domain.documentation
            node = self._search_xml(search_path, 'doc')
            if node:
                return strip_whitespace(node)
            node = self._search_xml(
                {'field': search_path['field']}, 'doc', True)
            if node:
                return strip_whitespace(node)
        LOGGER.warning('Could not find documentation for %r', search_path)

    @functools.lru_cache()
    def _domain(self, value: str) -> typing.Optional[Domain]:
        for domain in self._domains():
            if domain.name == value:
                return domain

    @functools.lru_cache()
    def _domains(self) -> typing.List[Domain]:
        values = []
        for value in self._amqp_json['domains']:
            domain = self._codegen_xml.xpath(
                '//amqp/domain[@name="{}"]'.format(value[0]))
            extension = self._extensions_xml.xpath(
                '//rabbitmq/domain[@name="{}"]'.format(value[0]))
            doc = self._codegen_xml.xpath(
                '//amqp/domain[@name="{}"]/doc'.format(value[0]))
            kwargs = {
                'name': value[0],
                'type': value[1],
                'documentation': self._remove_extra_whitespace(doc[0].text)
                if doc else None,
                'label': self._remove_extra_whitespace(
                    domain[0].attrib.get('label', '') if domain else ''),
                'nullable': True,
                'regex': None,
                'max_length': None,
                'default_value': domain[0].attrib.get('default-value', None)
                    if domain else None
            }
            if extension \
                    and extension[0].attrib.get('default-value') is not None:
                kwargs['default_value'] = \
                    extension[0].attrib.get('default-value')

            for assertion in self._codegen_xml.xpath(
                    '//amqp/domain[@name="{}"]/assert'.format(value[0])) or []:
                if assertion.attrib.get('check') == 'length':
                    kwargs['max_length'] = int(assertion.attrib['value'])
                elif assertion.attrib.get('check') == 'notnull':
                    kwargs['nullable'] = False
                elif assertion.attrib.get('check') == 'regexp':
                    kwargs['regex'] = assertion.attrib['value']

            for assertion in self._extensions_xml.xpath(
                    '//rabbitmq/domain[@name="{}"]/assert'.format(
                        value[0])) or []:
                if assertion.attrib.get('check') == 'length':
                    kwargs['max_length'] = int(assertion.attrib['value'])
                elif assertion.attrib.get('check') == 'notnull':
                    kwargs['nullable'] = False
                elif assertion.attrib.get('check') == 'regexp':
                    kwargs['regex'] = assertion.attrib['value']

            values.append(Domain(**kwargs))
        return values

    def _label(self, search_path: dict) -> typing.Optional[str]:
        """Attempt to return the short label documentation"""
        node = self._search_xml(search_path)
        if node and 'label' in node[0].attrib:
            return '{}{}'.format(
                node[0].attrib['label'][0:1].upper(),
                node[0].attrib['label'][1:])
        elif node and node[0].text:
            return '{}{}'.format(node[0].text.strip()[0:1].upper(),
                                 node[0].text.strip()[1:].strip())
        if 'field' in search_path:  # Look in domains and extensions
            domain = self._domain(search_path['field'])
            if domain and domain.label:
                return '{}{}'.format(domain.label[0:1].upper(),
                                     domain.label[1:].strip())
            node = self._search_xml(search_path, only_extensions=True)
            if node and 'label' in node[0].attrib:
                return '{}{}'.format(
                    node[0].attrib['label'][0:1].upper(),
                    node[0].attrib['label'][1:])
            node = self._search_xml({'field': search_path['field']})
            if node and 'label' in node[0].attrib:
                return '{}{}'.format(
                    node[0].attrib['label'][0:1].upper(),
                    node[0].attrib['label'][1:])

        LOGGER.warning('Could not find label for %r', search_path)

    @staticmethod
    def _load_codegen_json() -> dict:
        if not CODEGEN_JSON.exists():
            print('Downloading codegen JSON file to %s.' % CODEGEN_JSON)
            response = requests.get(CODEGEN_JSON_URL)
            if not response.ok:
                print('Error downloading JSON file: {}'.format(response))
                sys.exit(1)
            with CODEGEN_JSON.open('w') as handle:
                handle.write(response.content.decode('utf-8'))
        with CODEGEN_JSON.open('r') as handle:
            return json.load(handle)

    @staticmethod
    def _load_codegen_xml() -> lxml.etree.Element:
        if not CODEGEN_XML.exists():
            print('Downloading codegen XML file.')
            response = requests.get(CODEGEN_XML_URL)
            if not response.ok:
                print('Error downloading XML file: {}'.format(response))
                sys.exit(1)
            with CODEGEN_XML.open('w') as handle:
                handle.write(response.content.decode('utf-8'))

        with CODEGEN_XML.open('r') as handle:
            return lxml.etree.parse(handle).xpath('//amqp')[0]

    @staticmethod
    def _load_extensions_xml() -> lxml.etree.Element:
        with EXTENSIONS_XML.open('r') as handle:
            return lxml.etree.parse(handle).xpath('//rabbitmq')[0]

    def _lookup_field(self, class_name: str, method: str, field: str) -> dict:
        field_def = {}
        result = self._codegen_xml.xpath(
            '//amqp/class[@name="{}"]/method[@name="{}"]/'
            'field[@name="{}"]'.format(class_name, method, field))
        if result:
            for attribute in result[0].attrib:
                field_def[attribute] = result[0].attrib[attribute]
        result = self._extensions_xml.xpath(
            '//rabbitmq/class[@name="{}"]/method[@name="{}"]/'
            'field[@name="{}"]'.format(class_name, method, field))
        if result:
            for attribute in result[0].attrib:
                field_def[attribute] = result[0].attrib[attribute]
        return field_def

    @staticmethod
    def _pep8_class_name(value: str) -> str:
        """Returns a class name in the proper case per PEP8"""
        return_value = []
        parts = value.split('-')
        for part in parts:
            return_value.append(part[0:1].upper() + part[1:])
        return ''.join(return_value)

    @staticmethod
    def _remove_extra_whitespace(value: str) -> str:
        return ' '.join([dl.strip() for dl in value.split('\n')]).strip()

    def reset_output_buffer(self) -> None:
        self._output_buffer = []

    def _search_xml(self, search_path: dict,
                    suffix: typing.Optional[str] = None,
                    only_extensions: bool = False) \
            -> typing.Optional[lxml.etree.Element]:
        search = []
        for key in XPATH_ORDER:
            if key in search_path:
                search.append('%s[@name="%s"]' % (key, search_path[key]))
        if suffix:
            search.append(suffix)
        #  LOGGER.debug('Searching for %s', '/'.join(search))
        if only_extensions:
            return self._extensions_xml.xpath('/'.join(search))
        codegen = self._codegen_xml.xpath('/'.join(search))
        extension = self._extensions_xml.xpath('/'.join(search))
        if extension:
            codegen.extend(extension)
        return codegen

    def _write_file(self, path: pathlib.Path) -> None:
        if self.YAPF:
            code = yapf_api.FormatCode(
                '\n'.join(self._output_buffer), style_config='pep8')[0]
        else:
            code = '\n'.join(self._output_buffer)
        with path.open('w') as handle:
            handle.write(code)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    Codegen().build()



# -*- encoding: utf-8 -*-
import datetime
import unittest

from pamqp import decode, encode


class EncodeDecodeTests(unittest.TestCase):

    def test_encode_decode_field_table_long_keys(self):
        """Encoding and decoding a field_table with too long keys."""
        # second key is 126 A's + \N{PILE OF POO}
        data = {'A' * 256: 1,
                ((b'A' * 128) + b'\xf0\x9f\x92\xa9').decode('utf-8'): 2}
        encoded = encode.field_table(data)
        decoded = decode.field_table(encoded)[1]
        self.assertIn('A' * 128, decoded)

    def test_timestamp_with_dst(self):
        # this test assumes the system is set up using a northern hemisphere
        # timesone with DST (America/New_York as per github CI is fine)
        data = datetime.datetime(2006, 5, 21, 16, 30, 10,
                                 tzinfo=datetime.timezone.utc)
        encoded = encode.timestamp(data)
        decoded = decode.timestamp(encoded)[1]
        self.assertEqual(decoded, data)

    def test_timestamp_without_timezone(self):
        naive = datetime.datetime(2006, 5, 21, 16, 30, 10)
        aware = datetime.datetime(2006, 5, 21, 16, 30, 10,
                                  tzinfo=datetime.timezone.utc)
        encoded = encode.timestamp(naive)
        decoded = decode.timestamp(encoded)[1]
        self.assertEqual(decoded, aware)



import datetime

import pkg_resources
import sphinx_material

html_theme = 'sphinx_material'
html_theme_path = sphinx_material.html_theme_path()
html_context = sphinx_material.get_html_context()
html_sidebars = {
    "**": ["globaltoc.html", "searchbox.html"]
}
html_theme_options = {
    'base_url': 'http://pamqp.readthedocs.io',
    'repo_url': 'https://github.com/gmr/pamqp/',
    'repo_name': 'pamqp',
    'html_minify': True,
    'css_minify': True,
    'nav_title': 'pamqp',
    'globaltoc_depth': 1,
    'theme_color': 'fc6600',
    'color_primary': 'grey',
    'color_accent': 'orange',
    'version_dropdown': False
}
html_static_path = ['_static']
html_css_files = [
    'css/custom.css'
]

master_doc = 'index'
project = 'pamqp'
release = version = pkg_resources.get_distribution(project).version
copyright = '2011-{}, Gavin M. Roy'.format(datetime.date.today().year)

extensions = [
    'sphinx.ext.autodoc',
    'sphinx_autodoc_typehints',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx_material'
]

set_type_checking_flag = True
typehints_fully_qualified = True
always_document_param_types = True
typehints_document_rtype = True

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}

autodoc_default_options = {
    'autodoc_typehints': 'description',
    'special-members': ('__contains__,__eq__,__getitem__,'
                        '__iter__,__len__'),
}



# -*- encoding: utf-8 -*-
import datetime
import decimal
import struct
import unittest

from pamqp import decode

PLATFORM_32BIT = (struct.calcsize('P') * 8) == 32
PLATFORM_64BIT = (struct.calcsize('P') * 8) == 64


class CodecDecodeTests(unittest.TestCase):
    FIELD_ARR = (b'\x00\x00\x009b\x01u\xaf\xc8S\x00\x00\x00\x08Test \xe2\x9c'
                 b'\x88T\x00\x00\x00\x00Ec)\x92I\xbb\x9a\xca\x00D\x02\x00\x00'
                 b'\x01:f@H\xf5\xc3i\xc4e5\xffl\x80\x00\x00\x00\x00\x00\x00'
                 b'\x08')
    FIELD_ARR_VALUE = [
        1, 45000, 'Test ✈',
        datetime.datetime(2006, 11, 21, 16, 30, 10,
                          tzinfo=datetime.timezone.utc), -1147483648,
        decimal.Decimal('3.14'), 3.14, 3294967295, -9223372036854775800
    ]
    FIELD_TBL = (
        b'\x00\x00\x00\x99\x08arrayvalA\x00\x00\x00\x06b\x01b\x02b\x03\x07'
        b'boolvalt\x01\x06decvalD\x02\x00\x00\x01:\x07dictvalF\x00\x00\x00'
        b'\r\x04f\xe2\x9c\x89S\x00\x00\x00\x03\xe2\x9c\x90\x08floatvalf@H'
        b'\xf5\xc3\x06intvalb\x01\x07longvalI6e&U\x06strvalS\x00\x00\x00\x08'
        b'Test \xe2\x9c\x88\x0ctimestampvalT\x00\x00\x00\x00Ec)\x92\x04\xf0'
        b'\x9f\x90\xb0V'
    )
    FIELD_TBL_VALUE = {
        'intval': 1,
        'strval': 'Test ✈',
        'boolval': True,
        'timestampval': datetime.datetime(2006, 11, 21, 16, 30, 10,
                                          tzinfo=datetime.timezone.utc),
        'decval': decimal.Decimal('3.14'),
        '🐰': None,
        'floatval': 3.14,
        'longval': 912598613,
        'dictval': {
            'f✉': '✐'
        },
        'arrayval': [1, 2, 3]
    }

    def test_decode_by_type_invalid_data_type(self):
        self.assertRaises(ValueError, decode.by_type, b'Z\x00', b'foobar')

    def test_decode_bit_bytes_consumed(self):
        self.assertEqual(decode.bit(b'\xff', 4)[0], 0)

    def test_decode_invalid_value(self):
        self.assertRaises(ValueError, decode.bit, b'\xff', None)

    def test_decode_bit_on(self):
        self.assertTrue(decode.bit(b'\xff', 4)[1])

    def test_decode_bit_off(self):
        self.assertFalse(decode.bit(b'\x0f', 4)[1])

    def test_decode_boolean_bytes_consumed(self):
        self.assertEqual(decode.boolean(b'\x01')[0], 1)

    def test_decode_boolean_false(self):
        self.assertFalse(decode.boolean(b'\x00')[1])

    def test_decode_boolean_false_data_type(self):
        self.assertIsInstance(decode.boolean(b'\x00')[1], bool)

    def test_decode_boolean_invalid_value(self):
        self.assertRaises(ValueError, decode.boolean, None)

    def test_decode_boolean_true(self):
        self.assertTrue(decode.boolean(b'\x01')[1])

    def test_decode_boolean_true_data_type(self):
        self.assertIsInstance(decode.boolean(b'\x01')[1], bool)

    def test_decode_byte_array(self):
        self.assertEqual(decode.byte_array(b'\x00\x00\x00\x03ABC'),
                         (7, bytearray([65, 66, 67])))

    def test_decode_byte_array_invalid_value(self):
        self.assertRaises(ValueError, decode.byte_array, None)

    def test_decode_decimal_value_bytes_consumed(self):
        value = b'\x05\x00\x04\xcb/'
        self.assertEqual(decode.decimal(value)[0], len(value))

    def test_decode_decimal_value_data_type(self):
        value = b'\x05\x00\x04\xcb/'
        self.assertIsInstance(decode.decimal(value)[1], decimal.Decimal)

    def test_decode_decimal_value(self):
        value = b'\x05\x00\x04\xcb/'
        self.assertEqual(round(float(decode.decimal(value)[1]), 5),
                         round(float(decimal.Decimal('3.14159')), 5))

    def test_decode_decimal_invalid_value(self):
        self.assertRaises(ValueError, decode.decimal, False)

    def test_decode_double_value(self):
        value = b'@\t!\xf9\xf0\x1b\x86n'
        self.assertEqual(round(decode.double(value)[1], 5),
                         round(float(3.14159), 5))

    def test_decode_double_invalid_value(self):
        self.assertRaises(ValueError, decode.double, 123)

    def test_decode_embedded_value_null(self):
        self.assertEqual(decode.embedded_value(b'\00')[1], None)

    def test_decode_embedded_value_invalid_data(self):
        self.assertRaises(ValueError, decode.embedded_value, b'Z\x00')

    def test_decode_floating_point_bytes_consumed(self):
        value = b'@I\x0f\xd0'
        self.assertEqual(decode.floating_point(value)[0], 4)

    def test_decode_floating_point_data_type(self):
        value = b'@I\x0f\xd0'
        self.assertIsInstance(decode.floating_point(value)[1], float)

    def test_decode_floating_point_invalid_value(self):
        self.assertRaises(ValueError, decode.floating_point, False)

    def test_decode_floating_point_value(self):
        value = b'@I\x0f\xd0'
        self.assertEqual(round(decode.floating_point(value)[1], 5),
                         round(float(3.14159), 5))

    def test_decode_long_int_bytes_consumed(self):
        value = b'\x7f\xff\xff\xff'
        self.assertEqual(decode.long_int(value)[0], 4)

    def test_decode_long_int_data_type(self):
        value = b'\x7f\xff\xff\xff'
        self.assertIsInstance(decode.long_int(value)[1], int)

    def test_decode_long_int_invalid_value(self):
        self.assertRaises(ValueError, decode.long_int, None)

    def test_decode_long_int_value(self):
        value = b'\x7f\xff\xff\xff'
        self.assertEqual(decode.long_int(value)[1], 2147483647)

    def test_decode_long_long_int_bytes_consumed(self):
        value = b'\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertEqual(decode.long_long_int(value)[0], 8)

    @unittest.skipIf(PLATFORM_32BIT, 'Skipped on 32-bit platforms')
    def test_decode_long_long_int_data_type_64bit(self):
        value = b'\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertIsInstance(decode.long_long_int(value)[1], int)

    @unittest.skipIf(PLATFORM_64BIT, 'Skipped on 64-bit platforms')
    def test_decode_long_long_int_data_type_32bit(self):
        value = b'\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertIsInstance(decode.long_long_int(value)[1], int)

    def test_decode_long_long_int_invalid_value(self):
        self.assertRaises(ValueError, decode.long_long_int, None)

    def test_decode_long_long_int_value(self):
        value = b'\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertEqual(decode.long_long_int(value)[1], 9223372036854775800)

    def test_decode_long_str_bytes_consumed(self):
        value = b'\x00\x00\x00\n0123456789'
        self.assertEqual(decode.long_str(value)[0], 14)

    def test_decode_long_str_data_type(self):
        value = b'\x00\x00\x00\n0123456789'
        self.assertIsInstance(decode.long_str(value)[1], str)

    def test_decode_long_str_data_type_unicode(self):
        value = b'\0\0\0\x0c\xd8\xa7\xd8\xae\xd8\xaa\xd8\xa8\xd8\xa7\xd8\xb1'
        self.assertIsInstance(decode.long_str(value)[1], str)

    def test_decode_long_str_data_type_non_unicode(self):
        value = b'\x00\x00\x00\x01\xff'
        self.assertIsInstance(decode.long_str(value)[1], bytes)

    def test_decode_long_str_invalid_value(self):
        self.assertRaises(ValueError, decode.long_str, None)

    def test_decode_long_str_value(self):
        value = b'\x00\x00\x00\n0123456789'
        self.assertEqual(decode.long_str(value)[1], '0123456789')

    def test_decode_octet_bytes_consumed(self):
        value = b'\xff'
        self.assertEqual(decode.octet(value)[0], 1)

    def test_decode_octet_data_type(self):
        value = b'\xff'
        self.assertIsInstance(decode.octet(value)[1], int)

    def test_decode_octet_invalid_value(self):
        self.assertRaises(ValueError, decode.octet, None)

    def test_decode_octet_value(self):
        value = b'\xff'
        self.assertEqual(decode.octet(value)[1], 255)

    def test_decode_short_int_bytes_consumed(self):
        value = b'\x7f\xff'
        self.assertEqual(decode.short_int(value)[0], 2)

    def test_decode_short_int_data_type(self):
        value = b'\x7f\xff'
        self.assertIsInstance(decode.short_int(value)[0], int)

    def test_decode_short_int_invalid_value(self):
        self.assertRaises(ValueError, decode.short_int, None)

    def test_decode_short_int_value(self):
        value = b'\x7f\xff'
        self.assertEqual(decode.short_int(value)[1], 32767)

    def test_decode_short_short_int_data_type(self):
        self.assertIsInstance(decode.short_short_int(b'\xff')[0], int)

    def test_decode_short_short_int_invalid_value(self):
        self.assertRaises(ValueError, decode.short_short_int, None)

    def test_decode_short_short_uint_value(self):
        self.assertEqual(decode.short_short_uint(b'\xff')[1], 255)

    def test_decode_short_short_uint_invalid_value(self):
        self.assertRaises(ValueError, decode.short_short_uint, None)

    def test_decode_short_str_bytes_consumed(self):
        self.assertEqual(decode.short_str(b'\n0123456789')[0], 11)

    def test_decode_short_str_data_type(self):
        self.assertIsInstance(decode.short_str(b'\n0123456789')[1], str)

    def test_decode_short_str_invalid_value(self):
        self.assertRaises(ValueError, decode.short_str, None)

    def test_decode_short_str_value(self):
        self.assertEqual(decode.short_str(b'\n0123456789')[1], '0123456789')

    def test_decode_timestamp_bytes_consumed(self):
        self.assertEqual(decode.timestamp(b'\x00\x00\x00\x00Ec)\x92')[0], 8)

    def test_decode_timestamp_data_type(self):
        self.assertIsInstance(
            decode.timestamp(b'\x00\x00\x00\x00Ec)\x92')[1], datetime.datetime)

    def test_decode_timestamp_invalid_value(self):
        self.assertRaises(ValueError, decode.timestamp, None)

    def test_decode_timestamp_value(self):
        self.assertEqual(
            decode.timestamp(b'\x00\x00\x00\x00Ec)\x92')[1],
            datetime.datetime(2006, 11, 21, 16, 30, 10,
                              tzinfo=datetime.timezone.utc))

    def test_decode_field_array_bytes_consumed(self):
        self.assertEqual(
            decode.field_array(self.FIELD_ARR)[0], len(self.FIELD_ARR))

    def test_decode_field_array_data_type(self):
        self.assertIsInstance(decode.field_array(self.FIELD_ARR)[1], list)

    def test_decode_field_array_invalid_value(self):
        self.assertRaises(ValueError, decode.field_array, None)

    def test_decode_field_array_value(self):
        value = decode.field_array(self.FIELD_ARR)[1]
        for position in range(0, len(value)):
            if isinstance(value[position], float):
                self.assertAlmostEqual(
                    round(value[position], 3),
                    round(self.FIELD_ARR_VALUE[position], 3))
            else:
                self.assertEqual(value[position],
                                 self.FIELD_ARR_VALUE[position])

    def test_decode_field_table_bytes_consumed(self):
        self.assertEqual(
            decode.field_table(self.FIELD_TBL)[0], len(self.FIELD_TBL))

    def test_decode_field_table_data_type(self):
        self.assertIsInstance(decode.field_table(self.FIELD_TBL)[1], dict)

    def test_decode_field_table_invalid_value(self):
        self.assertRaises(ValueError, decode.field_table, None)

    def test_decode_field_table_value(self):
        value = decode.field_table(self.FIELD_TBL)[1]
        for key in self.FIELD_TBL_VALUE.keys():
            if isinstance(value[key], float):
                self.assertAlmostEqual(round(value[key], 3),
                                       round(self.FIELD_TBL_VALUE[key], 3))
            else:
                self.assertEqual(value[key], self.FIELD_TBL_VALUE[key])

    def test_decode_by_type_bit_bytes_consumed(self):
        self.assertEqual(decode.by_type(b'\xff', 'bit', 4)[0], 0)

    def test_decode_by_type_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, b'\xff', 'bit', None)

    def test_decode_by_type_bit_on(self):
        self.assertTrue(decode.by_type(b'\xff', 'bit', 4)[1])

    def test_decode_by_type_bit_off(self):
        self.assertFalse(decode.by_type(b'\x0f', 'bit', 4)[1])

    def test_decode_by_type_boolean_bytes_consumed(self):
        self.assertEqual(decode.by_type(b'\x01', 'boolean')[0], 1)

    def test_decode_by_type_boolean_false(self):
        self.assertFalse(decode.by_type(b'\x00', 'boolean')[1])

    def test_decode_by_type_boolean_false_data_type(self):
        self.assertIsInstance(decode.by_type(b'\x00', 'boolean')[1], bool)

    def test_decode_by_type_boolean_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, None, 'boolean')

    def test_decode_by_type_boolean_true(self):
        self.assertTrue(decode.by_type(b'\x01', 'boolean')[1])

    def test_decode_by_type_boolean_true_data_type(self):
        self.assertIsInstance(decode.by_type(b'\x01', 'boolean')[1], bool)

    def test_decode_by_type_byte_array_bytes_consumed(self):
        value = b'\x00\x00\x00\t123456789'
        self.assertEqual(decode.by_type(value, 'byte_array')[0], 13)

    def test_decode_by_type_byte_array_data_type(self):
        value = b'\x00\x00\x00\t123456789'
        self.assertIsInstance(
            decode.by_type(value, 'byte_array')[1], bytearray)

    def test_decode_by_type_byte_array_value(self):
        value = b'\x00\x00\x00\t123456789'
        self.assertEqual(
            decode.by_type(value, 'byte_array')[1], bytearray(b'123456789'))

    def test_decode_by_type_decimal_bytes_consumed(self):
        value = b'\x05\x00\x04\xcb/'
        self.assertEqual(decode.by_type(value, 'decimal')[0], len(value))

    def test_decode_by_type_decimal_data_type(self):
        self.assertIsInstance(
            decode.by_type(b'\x05\x00\x04\xcb/', 'decimal')[1],
            decimal.Decimal)

    def test_decode_by_type_decimal_value(self):
        self.assertEqual(
            round(float(decode.by_type(b'\x05\x00\x04\xcb/', 'decimal')[1]),
                  5), round(float(decimal.Decimal('3.14159')), 5))

    def test_decode_by_type_decimal_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, False, 'decimal')

    def test_decode_by_type_double_data_type(self):
        value = b'C\x0f\xd8\x91\x14\xb9\xc3\x98'
        self.assertIsInstance(decode.by_type(value, 'double')[1], float)

    def test_decode_by_type_double_bytes_consumed(self):
        value = b'C\x0f\xd8\x91\x14\xb9\xc3\x98'
        self.assertEqual(decode.by_type(value, 'double')[0], 8)

    def test_decode_by_type_double_value(self):
        value = b'C\x0f\xd8\x91\x14\xb9\xc3\x98'
        self.assertEqual(
            decode.by_type(value, 'double')[1], 1120480238450803.0)

    def test_decode_by_type_floating_point_data_type(self):
        self.assertIsInstance(decode.by_type(b'@I\x0f\xd0', 'float')[1], float)

    def test_decode_by_type_float_bytes_consumed(self):
        self.assertEqual(decode.by_type(b'@I\x0f\xd0', 'float')[0], 4)

    def test_decode_by_type_floating_point_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, False, 'float')

    def test_decode_by_type_floating_point_value(self):
        value = b'@I\x0f\xd0'
        self.assertEqual(round(decode.by_type(value, 'float')[1], 5),
                         round(float(3.14159), 5))

    def test_decode_by_type_long_bytes_consumed(self):
        value = b'\x7f\xff\xff\xff'
        self.assertEqual(decode.by_type(value, 'long')[0], 4)

    def test_decode_by_type_long_data_type(self):
        value = b'\x7f\xff\xff\xff'
        self.assertIsInstance(decode.by_type(value, 'long')[1], int)

    def test_decode_by_type_long_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, None, 'long')

    def test_decode_by_type_long_value(self):
        value = b'\x7f\xff\xff\xff'
        self.assertEqual(decode.by_type(value, 'long')[1], 2147483647)

    def test_decode_by_type_long_long_bytes_consumed(self):
        value = b'\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertEqual(decode.by_type(value, 'longlong')[0], 8)

    @unittest.skipIf(PLATFORM_64BIT, 'Skipped on 64-bit platforms')
    def test_decode_by_type_long_long_data_type_32bit(self):
        value = b'\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertIsInstance(decode.by_type(value, 'longlong')[1], int)

    @unittest.skipIf(PLATFORM_32BIT, 'Skipped on 32-bit platforms')
    def test_decode_by_type_long_long_data_type_64bit(self):
        value = b'\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertIsInstance(decode.by_type(value, 'longlong')[1], int)

    def test_decode_by_type_long_long_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, None, 'longlong')

    def test_decode_by_type_long_long_value(self):
        value = b'\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertEqual(
            decode.by_type(value, 'longlong')[1], 9223372036854775800)

    def test_decode_by_type_longstr_bytes_consumed(self):
        value = b'\x00\x00\x00\n0123456789'
        self.assertEqual(decode.by_type(value, 'longstr')[0], 14)

    def test_decode_by_type_longstr_data_type_with_unicode(self):
        value = b'\x00\x00\x00\x08Test \xe2\x9c\x88'
        self.assertIsInstance(decode.by_type(value, 'longstr')[1], str)

    def test_decode_by_type_longstr_data_type(self):
        value = b'\x00\x00\x00\n0123456789'
        self.assertIsInstance(decode.by_type(value, 'longstr')[1], str)

    def test_decode_by_type_longstr_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, None, 'longstr')

    def test_decode_by_type_longstr_value(self):
        value = b'\x00\x00\x00\n0123456789'
        self.assertEqual(decode.by_type(value, 'longstr')[1], '0123456789')

    def test_decode_by_type_octet_bytes_consumed(self):
        self.assertEqual(decode.by_type(b'\xff', 'octet')[0], 1)

    def test_decode_by_type_octet_data_type(self):
        self.assertIsInstance(decode.by_type(b'\xff', 'octet')[1], int)

    def test_decode_by_type_octet_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, None, 'octet')

    def test_decode_by_type_octet_value(self):
        self.assertEqual(decode.by_type(b'\xff', 'octet')[1], 255)

    def test_decode_by_type_short_bytes_consumed(self):
        value = b'\x7f\xff'
        self.assertEqual(decode.by_type(value, 'short')[0], 2)

    def test_decode_by_type_short_data_type(self):
        value = b'\x7f\xff'
        self.assertIsInstance(decode.by_type(value, 'short')[1], int)

    def test_decode_by_type_short_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, None, 'short')

    def test_decode_by_type_short_value(self):
        value = b'\x7f\xff'
        self.assertEqual(decode.by_type(value, 'short')[1], 32767)

    def test_decode_by_type_timestamp_bytes_consumed(self):
        value = b'\x00\x00\x00\x00Ec)\x92'
        self.assertEqual(decode.by_type(value, 'timestamp')[0], 8)

    def test_decode_by_type_timestamp_data_type(self):
        value = b'\x00\x00\x00\x00Ec)\x92'
        self.assertIsInstance(
            decode.by_type(value, 'timestamp')[1], datetime.datetime)

    def test_decode_by_type_timestamp_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, None, 'timestamp')

    def test_decode_by_type_timestamp_value(self):
        value = b'\x00\x00\x00\x00Ec)\x92'
        self.assertEqual(
            decode.by_type(value, 'timestamp')[1],
            datetime.datetime(2006, 11, 21, 16, 30, 10,
                              tzinfo=datetime.timezone.utc))

    def test_decode_by_type_void(self):
        self.assertIsNone(decode.by_type(b'', 'void')[1])

    def test_decode_by_type_field_array_bytes_consumed(self):
        self.assertEqual(
            decode.by_type(self.FIELD_ARR, 'array')[0], len(self.FIELD_ARR))

    def test_decode_by_type_field_array_data_type(self):
        self.assertIsInstance(decode.by_type(self.FIELD_ARR, 'array')[1], list)

    def test_decode_by_type_field_array_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, None, 'array')

    def test_decode_by_type_field_array_value(self):
        value = decode.by_type(self.FIELD_ARR, 'array')[1]
        for position in range(0, len(value)):
            if isinstance(value[position], float):
                self.assertAlmostEqual(
                    round(value[position], 3),
                    round(self.FIELD_ARR_VALUE[position], 3))
            else:
                self.assertEqual(
                    value[position], self.FIELD_ARR_VALUE[position])

    def test_decode_by_type_field_table_bytes_consumed(self):
        self.assertEqual(
            decode.by_type(self.FIELD_TBL, 'table')[0], len(self.FIELD_TBL))

    def test_decode_by_type_field_table_data_type(self):
        self.assertIsInstance(decode.by_type(self.FIELD_TBL, 'table')[1], dict)

    def test_decode_by_type_field_table_invalid_value(self):
        self.assertRaises(ValueError, decode.by_type, None, 'table')

    def test_decode_by_type_field_table_value(self):
        value = decode.by_type(self.FIELD_TBL, 'table')[1]
        for key in self.FIELD_TBL_VALUE.keys():
            if isinstance(value[key], float):
                self.assertAlmostEqual(
                    round(value[key], 3), round(self.FIELD_TBL_VALUE[key], 3))
            else:
                self.assertEqual(value[key], self.FIELD_TBL_VALUE[key])

    def test_decode_embedded_value_empty_bytes_consumed(self):
        self.assertEqual(decode.embedded_value(b'')[0], 0)

    def test_decode_embedded_value_empty_value(self):
        self.assertEqual(decode.embedded_value(b'')[1], None)

    def test_decode_embedded_value_decimal_bytes_consumed(self):
        value = b'D\x05\x00\x04\xcb/'
        self.assertEqual(decode.embedded_value(value)[0], len(value))

    def test_decode_embedded_value_decimal_data_type(self):
        value = b'D\x05\x00\x04\xcb/'
        self.assertIsInstance(
            decode.embedded_value(value)[1], decimal.Decimal)

    def test_decode_embedded_value_decimal_value(self):
        value = b'D\x05\x00\x04\xcb/'
        self.assertEqual(round(float(decode.embedded_value(value)[1]), 5),
                         round(float(decimal.Decimal('3.14159')), 5))

    def test_decode_embedded_value_double_bytes_consumed(self):
        value = b'dC\x0f\xd8\x91\x14\xb9\xc3\x98'
        self.assertEqual(decode.embedded_value(value)[0], len(value))

    def test_decode_embedded_value_double_data_type(self):
        value = b'dC\x0f\xd8\x91\x14\xb9\xc3\x98'
        self.assertIsInstance(decode.embedded_value(value)[1], float)

    def test_decode_embedded_value_double_value(self):
        value = b'dC\x0f\xd8\x91\x14\xb9\xc3\x98'
        self.assertEqual(decode.embedded_value(value)[1], 1120480238450803.0)

    def test_decode_embedded_value_long_bytes_consumed(self):
        value = b'I\x7f\xff\xff\xff'
        self.assertEqual(decode.embedded_value(value)[0], 5)

    def test_decode_embedded_value_long_data_type(self):
        value = b'I\x7f\xff\xff\xff'
        self.assertIsInstance(decode.embedded_value(value)[1], int)

    def test_decode_embedded_value_long_value(self):
        value = b'I\x7f\xff\xff\xff'
        self.assertEqual(decode.embedded_value(value)[1], 2147483647)

    def test_decode_embedded_value_long_uint_bytes_consumed(self):
        value = b'i\xff\xff\xff\xff'
        self.assertEqual(decode.embedded_value(value)[0], 5)

    def test_decode_embedded_value_long_uint_data_type(self):
        value = b'i\xff\xff\xff\xff'
        self.assertIsInstance(decode.embedded_value(value)[1], int)

    def test_decode_embedded_value_long_uint_value(self):
        value = b'i\xff\xff\xff\xff'
        self.assertEqual(decode.embedded_value(value)[1], 4294967295)

    def test_decode_embedded_value_long_long_bytes_consumed(self):
        value = b'l\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertEqual(decode.embedded_value(value)[0], 9)

    @unittest.skipIf(PLATFORM_32BIT, 'Skipped on 32-bit platforms')
    def test_decode_embedded_value_long_long_data_type_64bit(self):
        value = b'l\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertIsInstance(decode.embedded_value(value)[1], int)

    @unittest.skipIf(PLATFORM_64BIT, 'Skipped on 64-bit platforms')
    def test_decode_embedded_value_long_long_data_type_32bit(self):
        value = b'l\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertIsInstance(decode.embedded_value(value)[1], int)

    def test_decode_embedded_value_long_long_value(self):
        value = b'l\x7f\xff\xff\xff\xff\xff\xff\xf8'
        self.assertEqual(decode.embedded_value(value)[1], 9223372036854775800)

    def test_decode_embedded_value_longstr_bytes_consumed(self):
        value = b'S\x00\x00\x00\n0123456789'
        self.assertEqual(decode.embedded_value(value)[0], 15)

    def test_decode_embedded_value_longstr_data_type(self):
        value = b'S\x00\x00\x00\n0123456789'
        self.assertIsInstance(decode.embedded_value(value)[1], str)

    def test_decode_embedded_value_byte_array_data(self):
        value = b'x\x00\x00\x00\x03ABC'
        self.assertEqual(
            decode.embedded_value(value)[1], bytearray([65, 66, 67]))

    def test_decode_embedded_value_byte_array_data_type(self):
        value = b'x\x00\x00\x00\x03ABC'
        self.assertIsInstance(decode.embedded_value(value)[1], bytearray)

    def test_decode_embedded_value_longstr_data_type_unicode(self):
        value = (b'S\x00\x00\x00\x0c\xd8\xa7\xd8\xae\xd8\xaa\xd8\xa8\xd8'
                 b'\xa7\xd8\xb1')
        self.assertIsInstance(decode.embedded_value(value)[1], str)

    def test_decode_embedded_value_longstr_value(self):
        value = b'S\x00\x00\x00\n0123456789'
        self.assertEqual(decode.embedded_value(value)[1], '0123456789')

    def test_decode_embedded_value_short_bytes_consumed(self):
        value = b's\x7f\xff'
        self.assertEqual(decode.embedded_value(value)[0], 3)

    def test_decode_embedded_value_short_short_bytes_consumed(self):
        self.assertEqual(decode.embedded_value(b'b\xff')[0], 2)

    def test_decode_embedded_value_short_short_data_type(self):
        self.assertIsInstance(decode.embedded_value(b'b\xff')[1], int)

    def test_decode_embedded_value_short_short_value(self):
        self.assertEqual(decode.embedded_value(b'B\xff')[1], 255)

    def test_decode_embedded_value_short_data_type(self):
        value = b's\x7f\xff'
        self.assertIsInstance(decode.embedded_value(value)[1], int)

    def test_decode_embedded_value_short_value(self):
        self.assertEqual(decode.embedded_value(b's\x7f\xff')[1], 32767)

    def test_decode_embedded_value_short_uint_data_type(self):
        self.assertIsInstance(decode.embedded_value(b'u\xff\xff')[1], int)

    def test_decode_embedded_value_short_uint_value(self):
        self.assertEqual(decode.embedded_value(b'u\xff\xff')[1], 65535)

    def test_decode_embedded_value_timestamp_bytes_consumed(self):
        value = b'T\x00\x00\x00\x00Ec)\x92'
        self.assertEqual(decode.embedded_value(value)[0], 9)

    def test_decode_embedded_value_timestamp_data_type(self):
        value = b'T\x00\x00\x00\x00Ec)\x92'
        self.assertIsInstance(
            decode.embedded_value(value)[1], datetime.datetime)

    def test_decode_embedded_value_timestamp_value(self):
        value = b'T\x00\x00\x00\x00Ec)\x92'
        self.assertEqual(
            decode.embedded_value(value)[1],
            datetime.datetime(2006, 11, 21, 16, 30, 10,
                              tzinfo=datetime.timezone.utc))

    def test_decode_embedded_value_field_array_bytes_consumed(self):
        self.assertEqual(
            decode.embedded_value(b'A' + self.FIELD_ARR)[0],
            len(b'A' + self.FIELD_ARR))

    def test_decode_embedded_value_field_array_data_type(self):
        self.assertIsInstance(
            decode.embedded_value(b'A' + self.FIELD_ARR)[1], list)

    def test_decode_embedded_value_field_array_value(self):
        value = decode.embedded_value(b'A' + self.FIELD_ARR)[1]
        for position in range(0, len(value)):
            if isinstance(value[position], float):
                self.assertAlmostEqual(
                    round(value[position], 3),
                    round(self.FIELD_ARR_VALUE[position], 3))
            else:
                self.assertEqual(value[position],
                                 self.FIELD_ARR_VALUE[position])

    def test_decode_embedded_value_field_table_bytes_consumed(self):
        self.assertEqual(
            decode.embedded_value(b'F' + self.FIELD_TBL)[0],
            len(b'F' + self.FIELD_TBL))

    def test_decode_embedded_value_field_table_data_type(self):
        self.assertIsInstance(
            decode.embedded_value(b'F' + self.FIELD_TBL)[1], dict)

    def test_decode_embedded_value_field_table_value(self):
        value = decode.embedded_value(b'F' + self.FIELD_TBL)[1]
        for key in self.FIELD_TBL_VALUE.keys():
            if isinstance(value[key], float):
                self.assertAlmostEqual(round(value[key], 3),
                                       round(self.FIELD_TBL_VALUE[key], 3))
            else:
                self.assertEqual(value[key], self.FIELD_TBL_VALUE[key])

    def test_decode_embedded_value_void_consumed(self):
        self.assertEqual(decode.embedded_value(b'V')[0], 1)

    def test_decode_embedded_value_void_value(self):
        self.assertIsNone(decode.embedded_value(b'V')[1])

    def test_field_embedded_value_field_table_keys(self):
        value = decode.embedded_value(b'F' + self.FIELD_TBL)[1]
        self.assertListEqual(sorted(value.keys()),
                             sorted(self.FIELD_TBL_VALUE.keys()))

    def test_decode_large_timestamp_bytes_consumed(self):
        dt = datetime.datetime(2107, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
        large_timestamp_bytes = struct.pack('>Q', int(dt.timestamp() * 1000))
        self.assertEqual(decode.timestamp(large_timestamp_bytes)[0], 8)

    def test_decode_large_timestamp_data_type(self):
        dt = datetime.datetime(2107, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
        large_timestamp_bytes = struct.pack('>Q', int(dt.timestamp() * 1000))
        self.assertIsInstance(decode.timestamp(large_timestamp_bytes)[1],
                              datetime.datetime)

    def test_decode_large_timestamp_value(self):
        dt = datetime.datetime(2107, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
        large_timestamp_bytes = struct.pack('>Q',
                                            int(dt.timestamp() * 1000))
        self.assertEqual(decode.timestamp(large_timestamp_bytes)[1], dt)



import unittest
import uuid

from pamqp import commands


class ArgumentErrorsTestCase(unittest.TestCase):

    def test_basic_consume_queue_length(self):
        with self.assertRaises(ValueError):
            commands.Basic.Consume(queue=str.ljust('A', 257))

    def test_basic_consume_queue_characters(self):
        with self.assertRaises(ValueError):
            commands.Basic.Consume(queue='*')

    def test_basic_consume_ticket(self):
        with self.assertRaises(ValueError):
            commands.Basic.Consume(ticket=46 & 2)  # Just ahead of me

    def test_basic_deliver_exchange_length(self):
        with self.assertRaises(ValueError):
            commands.Basic.Deliver('ctag0', 1, False, str.ljust('A', 128), 'k')

    def test_basic_deliver_exchange_characters(self):
        with self.assertRaises(ValueError):
            commands.Basic.Deliver('ctag0', 1, False, '*', 'k')

    def test_basic_get_queue_length(self):
        with self.assertRaises(ValueError):
            commands.Basic.Get(queue=str.ljust('A', 257))

    def test_basic_get_queue_characters(self):
        with self.assertRaises(ValueError):
            commands.Basic.Get(queue='*')

    def test_get_ticket(self):
        with self.assertRaises(ValueError):
            commands.Basic.Get(ticket=46 & 2)

    def test_basic_getempty_cluster_id(self):
        with self.assertRaises(ValueError):
            commands.Basic.GetEmpty(cluster_id='See my shadow changing')

    def test_basic_getok_exchange_length(self):
        with self.assertRaises(ValueError):
            commands.Basic.GetOk(1, False, str.ljust('A', 128), 'k', 0)

    def test_basic_getok_exchange_characters(self):
        with self.assertRaises(ValueError):
            commands.Basic.GetOk(1, False, '*', 'k', 0)

    def test_basic_properties_cluster_id(self):
        with self.assertRaises(ValueError):
            commands.Basic.Properties(cluster_id='See my shadow changing')

    def test_basic_publish_exchange_length(self):
        with self.assertRaises(ValueError):
            commands.Basic.Publish(exchange=str.ljust('A', 128))

    def test_basic_publish_exchange_characters(self):
        with self.assertRaises(ValueError):
            commands.Basic.Publish(exchange='*')

    def test_basic_publish_ticket(self):
        with self.assertRaises(ValueError):
            commands.Basic.Publish(ticket=46 & 2)

    def test_basic_return_exchange_length(self):
        with self.assertRaises(ValueError):
            commands.Basic.Return(404, 'Not Found', str.ljust('A', 128), 'k')

    def test_basic_return_exchange_characters(self):
        with self.assertRaises(ValueError):
            commands.Basic.Return(404, 'Not Found', '*', 'k')

    def test_connection_open_vhost(self):
        with self.assertRaises(ValueError):
            commands.Connection.Open(str.ljust('A', 128))

    def test_connection_open_capabilities(self):
        with self.assertRaises(ValueError):
            commands.Connection.Open(capabilities=str(uuid.uuid4()))

    def test_connection_open_insist(self):
        with self.assertRaises(ValueError):
            commands.Connection.Open(insist=True)

    def test_connection_openok_known_hosts(self):
        with self.assertRaises(ValueError):
            commands.Connection.OpenOk(known_hosts=str(uuid.uuid4()))

    def test_channel_open_out_of_band(self):
        with self.assertRaises(ValueError):
            commands.Channel.Open(out_of_band=str(uuid.uuid4()))

    def test_channel_openok_channel_id(self):
        with self.assertRaises(ValueError):
            commands.Channel.OpenOk(channel_id=str(uuid.uuid4()))

    def test_exchange_declare_ticket(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Declare(ticket=46 & 2)  # Just ahead of me

    def test_exchange_declare_exchange_length(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Declare(exchange=str.ljust('A', 128))

    def test_exchange_declare_exchange_characters(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Declare(exchange='***')

    def test_exchange_delete_ticket(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Delete(ticket=46 & 2)

    def test_exchange_delete_exchange_length(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Delete(exchange=str.ljust('A', 128))

    def test_exchange_delete_exchange_characters(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Delete(exchange='***')

    def test_exchange_bind_ticket(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Bind(ticket=46 & 2)

    def test_exchange_bind_destination_length(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Bind(destination=''.rjust(32768, '*'))

    def test_exchange_bind_destination_pattern(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Bind(destination='Fortysix & 2')

    def test_exchange_bind_source_length(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Bind(source=''.rjust(32768, '*'))

    def test_exchange_bind_source_pattern(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Bind(source='Fortysix & 2')

    def test_exchange_unbind_ticket(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Unbind(ticket=46 & 2)

    def test_exchange_unbind_destination_length(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Unbind(destination=''.rjust(32768, '*'))

    def test_exchange_unbind_destination_pattern(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Unbind(destination='Fortysix & 2')

    def test_exchange_unbind_source_length(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Unbind(source=''.rjust(32768, '*'))

    def test_exchange_unbind_source_pattern(self):
        with self.assertRaises(ValueError):
            commands.Exchange.Unbind(source='Fortysix & 2')

    def test_queue_declare_ticket(self):
        with self.assertRaises(ValueError):
            commands.Queue.Declare(ticket=46 & 2)

    def test_queue_declare_queue_length(self):
        with self.assertRaises(ValueError):
            commands.Queue.Declare(queue=str.ljust('A', 257))

    def test_queue_declare_queue_characters(self):
        with self.assertRaises(ValueError):
            commands.Queue.Declare(queue='***')

    def test_queue_declare_queue_with_valid_name(self):
        self.assertIsNotNone(commands.Queue.Declare(queue='foo'))

    def test_queue_declareok_queue_length(self):
        with self.assertRaises(ValueError):
            commands.Queue.DeclareOk(str.ljust('A', 257), 0, 0)

    def test_queue_declareok_queue_characters(self):
        with self.assertRaises(ValueError):
            commands.Queue.DeclareOk('***', 0, 0)

    def test_queue_delete_ticket(self):
        with self.assertRaises(ValueError):
            commands.Queue.Delete(ticket=46 & 2)

    def test_queue_delete_queue_length(self):
        with self.assertRaises(ValueError):
            commands.Queue.Delete(queue=str.ljust('A', 257))

    def test_queue_delete_queue_characters(self):
        with self.assertRaises(ValueError):
            commands.Queue.Delete(queue='***')

    def test_queue_bind_ticket(self):
        with self.assertRaises(ValueError):
            commands.Queue.Bind(ticket=46 & 2)

    def test_queue_bind_queue_length(self):
        with self.assertRaises(ValueError):
            commands.Queue.Bind(queue=str.ljust('A', 257), exchange='B')

    def test_queue_bind_queue_characters(self):
        with self.assertRaises(ValueError):
            commands.Queue.Bind(queue='***', exchange='B')

    def test_queue_bind_exchange_length(self):
        with self.assertRaises(ValueError):
            commands.Queue.Bind(exchange=str.ljust('A', 128), queue='B')

    def test_queue_bind_exchange_characters(self):
        with self.assertRaises(ValueError):
            commands.Queue.Bind(exchange='***', queue='B')

    def test_queue_unbind_ticket(self):
        with self.assertRaises(ValueError):
            commands.Queue.Unbind(ticket=46 & 2)

    def test_queue_unbind_queue_length(self):
        with self.assertRaises(ValueError):
            commands.Queue.Unbind(queue=str.ljust('A', 257), exchange='B')

    def test_queue_unbind_queue_characters(self):
        with self.assertRaises(ValueError):
            commands.Queue.Unbind(queue='***', exchange='B')

    def test_queue_unbind_exchange_length(self):
        with self.assertRaises(ValueError):
            commands.Queue.Unbind(exchange=str.ljust('A', 128), queue='B')

    def test_queue_unbind_exchange_characters(self):
        with self.assertRaises(ValueError):
            commands.Queue.Unbind(exchange='***', queue='B')

    def test_queue_purge_ticket(self):
        with self.assertRaises(ValueError):
            commands.Queue.Purge(ticket=46 & 2)

    def test_queue_purge_queue_length(self):
        with self.assertRaises(ValueError):
            commands.Queue.Purge(queue=str.ljust('A', 257))

    def test_queue_purge_queue_characters(self):
        with self.assertRaises(ValueError):
            commands.Queue.Purge(queue='***')



import unittest

from pamqp import commands

# Tag uri examples from https://en.wikipedia.org/wiki/Tag_URI_scheme
TAG_URIS = [
    'tag:timothy@hpl.hp.com,2001:web/externalHome',
    'tag:sandro@w3.org,2004-05:Sandro',
    'tag:my-ids.com,2001-09-15:TimKindberg:presentations:UBath2004-05-19',
    'tag:blogger.com,1999:blog-555',
    'tag:yaml.org,2002:int#section1'
]


class TagUriScheme(unittest.TestCase):

    def test_tag_uri_scheme_tag1(self):
        commands.Exchange.Declare(exchange=TAG_URIS[0])

    def test_tag_uri_scheme_tag2(self):
        commands.Exchange.Declare(exchange=TAG_URIS[1])

    def test_tag_uri_scheme_tag3(self):
        commands.Exchange.Declare(exchange=TAG_URIS[2])

    def test_tag_uri_scheme_tag4(self):
        commands.Exchange.Declare(exchange=TAG_URIS[3])



# -*- encoding: utf-8 -*-
import datetime
import unittest
import uuid

from pamqp import body, commands, frame, header, heartbeat


class MarshalingTests(unittest.TestCase):
    def test_protocol_header(self):
        expectation = b'AMQP\x00\x00\t\x01'
        response = frame.marshal(header.ProtocolHeader(), 0)
        self.assertEqual(response, expectation,
                         'ProtocolHeader did not match expectation')

    def test_basic_ack(self):
        expectation = (b'\x01\x00\x01\x00\x00\x00\r\x00<\x00P\x00\x00\x00'
                       b'\x00\x00\x00\x00\x01\x00\xce')
        frame_obj = commands.Basic.Ack(1, False)
        response = frame.marshal(frame_obj, 1)
        self.assertEqual(response, expectation,
                         'Basic.Ack did not match expectation')

    def test_basic_cancel(self):
        expectation = (b'\x01\x00\x01\x00\x00\x00\r\x00<\x00\x1e\x07'
                       b'ctag1.0\x00\xce')
        frame_obj = commands.Basic.Cancel('ctag1.0', False)
        response = frame.marshal(frame_obj, 1)
        self.assertEqual(response, expectation,
                         'Basic.Cancel did not match expectation')

    def test_basic_cancelok(self):
        expectation = (b'\x01\x00\x01\x00\x00\x00\x0c\x00<\x00\x1f\x07'
                       b'ctag1.0\xce')
        frame_obj = commands.Basic.CancelOk(consumer_tag='ctag1.0')
        response = frame.marshal(frame_obj, 1)
        self.assertEqual(response, expectation,
                         'Basic.Cancel did not match expectation')

    def test_basic_consume(self):
        expectation = (b'\x01\x00\x01\x00\x00\x00\x15\x00<\x00\x14\x00'
                       b'\x00\x03bar\x05ctag0\x00\x00\x00\x00\x00\xce')
        frame_obj = commands.Basic.Consume(0, 'bar', 'ctag0', False,
                                           False, False, False)
        response = frame.marshal(frame_obj, 1)
        self.assertEqual(response, expectation,
                         'Basic.Consume did not match expectation')

    def test_heartbeat(self):
        expectation = b'\x08\x00\x00\x00\x00\x00\x00\xce'
        response = frame.marshal(heartbeat.Heartbeat(), 0)
        self.assertEqual(response, expectation,
                         'Heartbeat did not match expectation')

    def test_content_body(self):
        value = str(uuid.uuid4()).encode('utf-8')
        expectation = b'\x03\x00\x01\x00\x00\x00$' + value + b'\xce'
        self.assertEqual(frame.marshal(body.ContentBody(value), 1),
                         expectation)

    def test_content_header(self):
        expectation = (b'\x02\x00\x01\x00\x00\x00\x0e\x00<\x00\x00\x00\x00'
                       b'\x00\x00\x00\x00\x00\n\x00\x00\xce')
        self.assertEqual(frame.marshal(header.ContentHeader(body_size=10), 1),
                         expectation)

    def test_content_header_with_basic_properties(self):
        props = commands.Basic.Properties(
            app_id='unittest',
            content_type='application/json',
            content_encoding='bzip2',
            correlation_id='d146482a-42dd-4b8b-a620-63d62ef686f3',
            delivery_mode=2,
            expiration='100',
            headers={'foo': 'Test ✈'},
            message_id='4b5baed7-66e3-49da-bfe4-20a9651e0db4',
            message_type='foo',
            priority=10,
            reply_to='q1',
            timestamp=datetime.datetime(2019, 12, 19, 23, 29, 00,
                                        tzinfo=datetime.timezone.utc))
        expectation = (b'\x02\x00\x01\x00\x00\x00\xa2\x00<\x00\x00\x00\x00\x00'
                       b'\x00\x00\x00\x00\n\xff\xe8\x10application/json\x05'
                       b'bzip2\x00\x00\x00\x11\x03fooS\x00\x00\x00\x08Test '
                       b'\xe2\x9c\x88\x02\n$d146482a-42dd-4b8b-a620-63d62ef68'
                       b'6f3\x02q1\x03100$4b5baed7-66e3-49da-bfe4-20a9651e0db4'
                       b'\x00\x00\x00\x00]\xfc\x07\xbc\x03foo\x08unittest\xce')
        self.assertEqual(frame.marshal(header.ContentHeader(0, 10, props), 1),
                         expectation)

    def test_unknown_frame_type(self):
        with self.assertRaises(ValueError):
            frame.marshal(self, 1)
